local Effect = class:extend("Effect")

--what an effect can target/affect or spread through
local classes = {"DungeonRoom","Item","Creature","Entity"}


local rarities = {
    {1,30},
    {2,25},
    {3,20},
    {4,20},
    {5,15},
    {6,10},
    {7, 7},
    {8, 7},
    {9, 5},
    {10, 2}
}


local f = {"Creature"}
local e = {"Item"}
local d = {"Entity"}
local c = {"Creature","DungeonRoom"}
local b = {"Item","DungeonRoom"}
local a = {"Entity","DungeonRoom"}

local spread = {
    {f,e,e},
    {f,f,f,f,f,f,d,e,e,e},
    {f,f,f,f,f,f,d,d,d,d,d,d,e,e},
    {f,f,f,f,d,d,d,c,e,e,e,e,e,e,e,e,e,e,},
    {f,f,f,f,d,d,d,d,d,d,d,d,c,c,c,b,e,e},
    {f,f,b,b,b,b,b,b,b,b,c,c,c,c,c,c,c,d},
    {f,e,d},
    {f,e,d,c,b,a},
    {a,b,b,b,c,c,c,d},
    {a,a,a,a,d,d,d,d,c,c,c,b,f,e}
}

local isDisease = {
    20,
    20,
    30,
    40,
    20,
    70,
    50,
    15,
    10,
    40
}

local increment = {
    {.5,.6,.7,.5,.8,1},
    {1,1,1,1,1.5,1,2,.5},
    {1,1,1,2,1.5,1},
    {1,2,1.5,.25,1.25,1.25},
    {1,1.25,1.25,1.5,1.5,2},
    {1.25,1.5,1.5,1.6,1.6,1.8,2,2.3},
    {1.5,1.5,1.5,1.25,2,2.5},
    {1.5,1.5,1.5,2,2,2.25,3,2,2.5},
    {2,2,2,2,2,2,1.5,1.5,1.8,1.8,1.9,1.5,1.5,1.5,1.5,1.5,2.5},
    {2,2,2,2,2,2,2.5,2.5,2.5,2.5,2.5,3,3,5}
}

local i10
local ii10 = increment[10]

i10 = #increment[10]
for x = 1, i10 do
    increment[10][i10+x] = ii10[x]
end
increment[10][#increment[10]+1] = 10

i10 = #increment[10]
for x = 1, i10 do
    increment[10][i10+x] = ii10[x]
end

increment[10][#increment[10]+1] = 1000

Effect._diseases = {{},{},{},{},{},{},{},{},{},{}}
Effect._blessings = {{},{},{},{},{},{},{},{},{},{}}
Effect._curses = {{},{},{},{},{},{},{},{},{},{}}

function Effect.loadEffects()
    Effect._effects = {}
    Effect._rarities = {}
    
    for x, i in pairs(rarities) do
        for xx = 1, i[2] do
            Effect._rarities[#Effect._rarities+1] = i[1]
        end
    end
    
    for i = 1, 100 do
        local data = getValue(effects)
        local eff  = Effect:new({data=data})
        eff.rarity = eff.rarity+getValue(Effect._rarities[self.rarity])
        if eff.rarity > 10 then
            eff.rarity = 10
        end
        
        local e = eff
        e.increment = getValue(e.increment)-1+getValue(increment[e.rarity])
        e.timeIncrement = getValue(e.timeIncrement)-1+getValue(increment[e.rarity])
        e.isDisease = e.isDisease or (isDisease[e.rarity]*e:getNegativity(1.5))>=math.random(101)
        e.spreadThrough = getValue(spread[e.rarity])
        e.spreadRate = e.spreadRate or lume.max(20,lume.min(100, math.random(e.rarity*2*10)))
        
        if e.isDisease then
            table.add(Effect._diseases[e.rarity], e)
            
            if e:getPositivity() > 0 then
                table.add(Effect._blessings[e.rarity], e)
            else
                table.add(Effect._curses[e.rarity], e)
            end
        end
        
        e:getRealValues()
    end
        
end

function Effect:__init__(k)
    local data = lume.copy(k.data)
    
    self.data = data
    
    self:load(data)
end

function Effect:load(data)
    self.name = data.name
    
    
    self.verb = data.verb
    self.past = data.past or string.format("%sed",self.verb)
    selr.continuous = data.continuous or string.format("%sing",self.verb)
    
    self.targets = data.affects or data.targets
    self.targetsLen = #self.targets
    
    self.increment = data.increment or 1
    self.timeIncrement = data.timeIncrement or 1
    
    self.boost = lume.copy(data.boost or {})
    self.mix = data.mix or {}
    
    self.tags = data.tags
    
    self._update = data.update
    self._draw   = data.draw
end

function Effect:getNegativity(max)
    local max = max or 1
    
    local maxVal = 0
    local count = 0
    local val = 0
    for x, i in pairs(self.boost) do
        if i ~= 0 then
            val = val - i
            maxVal = lume.max(maxVal, math.abs(i))
            count = count+1
        end
    end
    
    val = ((val/count)/maxVal)*max
    
    return max
end

function Effect:getPositivity(max)
    return -self:getNegativity(max)
end

function Effect:clone()
    local eff = Effect:new({data=self.data})
    
    for x, i in pairs(self) do
        if type(i) == "table" and not (i.class) then
            eff[x] = lume.copy(i)
        else
            eff[x] = i
        end
    end
    
    return eff
end

-- spread rate in diseases could be the chance it has to spread or
-- time it takes to accumulate once spread?

function Effect:getRealValues()
    self.time = self.time*self.timeIncrement
    
    for x, i in pairs(self.boost) do
        self.boost[x] = i*self.increment
    end
    
    self.spreadRate = self.spreadRate*self.increment*math.random(100)/100
end

function Effect:affect(object)
    local continue = false
    for x = 1, self.targetsLen do
        local class = self.targets[x]
        continue = (object[string.format("is%s",class)] and self.canAffect[class])
        if continue then
            break
        end
    end
    
    if not continue then
        return false
    end
    
    object:addEffect(self)
    self.object = object
    
    return true
end

function Effect:play(obj)
    local obj = obj or self.object
    
    for name, value in pairs(self.boost) do
        local stat = obj[name]
        if stat then
            obj[name] = obj[name] + value
        end
    end
end

function Effect:parse(object)
    -- object can be an Entity (Item or Creature)  or a Dungeon Room
    -- all effects should be checked when
    ---- an entity comes into a room
    ---- a creature gets an item
    ---- a creature attacks an entity
end

function Effect:update(time)
    if self._update then
        self:_update(time)
    end
    
    self:play(self.object)
    
    self.time = self.time - time
    
    if self.time <= 0 then
        self.object:removeEffect(self)
    end
end

function Effect:draw()
    if self._draw then
        self:_draw()
    end
end

return Effect