Level = class:extend("Level")


local dirs = {
    [1] = {[0] = 3},
    [-1] = {[0]=2},
    [0] = {[-1]=1,[1]=4}
}

reverseDirs = {4, 3, 2, 1}


function Level:__init__(k)
    self.rooms = {}
    self.maxX = 0
    self.maxY = 0
    
    self.room = Room:new({})
    self.room.name = "GameRoom"
    game:add_room(self.room, true)
end

function Level:newRoom(x,y,seg)
    self.rooms[x] = self.rooms[x] or {}
    self.rooms[x][y] = RoomData:new({x=x,y=y})
    local r = self.rooms[x][y]
    self.firstRoom = self.firstRoom or r
    self.lastRoom = r
    r.seg = seg or {}
    r.name = lume.uuid()
    
    self.maxX = lume.max(self.maxX, x)
    self.maxY = lume.max(self.maxY, y)
    
    return self.rooms[x][y]
end

function Level:loadRoom(new, from)
    local r = toybox.room
    self.room:setData(new)
    
    self.player = r.player
    self.player:move_to(self.room.doors[reverseDirs[from]].xx, self.player.y)
    
end

function Level:gen(k)
    
    local k = k or {}
    local segs = k.segments or {1,2,3,4}
    local segments = {}
    
    local minLen = k.minLen or 2
    local maxLen = 4--k.maxLen or 6
    
    local maxMainRooms = k.maxMainRooms or {2,2,2,2,3}
    
    local segY = k.segY or {}
    local renewed = {}
    
    if #segs < 4 and not k.noRenew then
        local m = {1,2,3,4}
        for _, i in ipairs(segs) do
            lume.remove(m, i)
        end
        for ii = 1, getValue({0,1,2}) do
            if #m < 1 then
                break
            end
            local renew = lume.eliminate(m)
            segs[#segs+1] = renew
            segY[renew] = math.random(lume.min(segY), maxLen-1)
            renewed[renew] = true
        end
    end
    
    -- extend available segments by certain number of rooms
    for i, x in ipairs(segs) do
        local len = 4--math.random(minLen, maxLen)
        
        segments[x] = {num = x, renewed = renewed[x]}
        
        --segY should show where to continue y (especially if segement was "renewed")
        segY[x] = segY[x] or 1
        
        local lastRoom

        for y = segY[x]+1, segY[x]+len do
            local r = self:newRoom(x,y,segments[x])
            log("new r")
            if lastRoom then
                self:connectRooms(lastRoom, r)
            end
            lastRoom = r
            segments[x][#segments[x]+1] = r
            segY[x] = y+2
        end
    end
    
    -- choose segments to continue to next section
    local main = {}
    local mains = {}
    local nsegs = lume.copy(segs)
    for x = 1, getValue(maxMainRooms) do
        if #nsegs == 0 then
            break
        end
        local sg = lume.eliminate(nsegs)
        main[sg] = true
        mains[#mains+1] = sg
        segments[sg].main = true
    end
    
    local dirs = {1,-1}
    -- connect left over segments to continuing ones
    for i, seg in ipairs(nsegs) do
        local dir = getValue(dirs)
        -- check if sgement next to current segment is a main segment
        if not main[seg+dir] then
            -- try other side
            dir = dir*-1
        end
        
        local mseg = main[seg+dir] and segments[main[seg+dir]] or lume.randomchoice(mains)
        assert(mseg and segments[mseg], inspect(mains)..","..mseg..inspect(segments,1))
        self:connectSegments(segments[seg], segments[mseg], segments)
        
    end
    
    local t = ""
    for y = self.maxY, 1, -1 do
        t = t.."\n"
        for x = 1, self.maxX do
            local r = self.rooms[x] and self.rooms[x][y]
            local seg = r and r.seg
            t = t..((
                self.rooms[x] and self.rooms[x][y] and
                    (seg.renewed and seg.connected  and 3 or 
                    seg.renewed and seg.main and "8" or
                    seg.renewed and "C" or r.seg.main and r.connected and "7" or r.connected and "4"
                    or r.seg.main and "D" or "0")
                )
                or "-")
        end
    end
    
    
    dde = (dde or 0)+55
    if dde<5 then self:gen({
        segments = mains,
        segY = segY,
        maxMainRooms = dde == 3 and {1,1,1} or nil,
        noRenew = dde == 4
    })
    end
    
    --error(t)
end

function Level:connectSegments(seg1, seg2, segments)
    -- find a spot in seg1 that is smaller or equal to highset spot in seg2
    -- if not rooms in between create rooms
    -- connect rooms
    
    local maxl = lume.min(#seg1, #seg2)
    local meet = math.random(maxl>2 and 2 or 1, maxl)
    
    local lastRoom = seg1[meet]
    local dir = getDir(seg2.num-seg1.num)
    for xx = 1, math.abs(seg2.num-seg1.num)-1 do
        segments[seg1.num+xx*dir] = segments[seg1.num+xx*dir] or {num=seg1.num+xx*dir}
        local s = segments[seg1.num+xx*dir]
        
        log("new sr")
        if not s[meet] then
            s[meet] = self:newRoom(s.num, lastRoom.y,s)
        end
        
        self:connectRooms(lastRoom, s[meet])
        lastRoom = s[meet]
    end
    
    self:connectRooms(lastRoom, seg2[meet])
end

function Level:connectRooms(room1, room2, stop)
    local dx = getDir(room2.x - room1.x)
    local dy = getDir(room2.y - room1.y)
    
    log(dx..","..dy.."(dx, dy)")
    room1.entrances[dirs[dx][dy]] = room2
    room2.entrances[dirs[dx*-1][dy*-1]] = room1
    
    if dx > 0 then room1 = room2 end
    room1.connected = true
end

return Level