local Person = Thing:extend("Person")

function Person:create(k)
    game.set_image_filter = "nearest"
    
    Thing.create(self, k)
    
    self.legs = toybox.new_sprite(self, {
        animations = {
            idle = {
                source = "person/legs_stance",
                useImages = true,
                delay = .3,
                mode = "loop"
            },
            walk = {
                source = "person/legs_walk",
                delay = .3,
                mode = "loop",
                useImages = true
            }
        }
    }):switch("walk")
    
    self.body = toybox.new_sprite(self, {
        animations = {
            idle = {
                source = "person/body",
                useImages = true,
                delay = .3
            }
        }
    })
    
    self.left_arm = toybox.new_sprite(self, {
        animations = {
            idle = {
                source = "person/left_arm",
                useImages = true,
                delay = .3,
                mode = "loop"
            }
        }
    })
    
    self.right_arm = toybox.new_sprite(self, {
        animations = {
            idle = {
                source = "person/right_arm",
                useImages = true,
                delay = .3,
                mode = "loop"
            }
        }
    })
    
    
    self:add_sprite(self.left_arm)
    self:add_sprite(self.legs)
    self:add_sprite(self.body)
    self:add_sprite(self.right_arm)
    
    self.head = self:add_image("person/head.png")
    self.head._w , self.head._h = resizeImage(self.head.source, self.w,self.h)
    self:set_box(nil,nil,self.w/4,self.h*.7)
    self:center()
    --self:offset(nil,self.offset_y+40*rhh)
    
    self.hair_base = self:add_image("none.png")
    self.hair_outline = self:add_image("none.png")
    self.exHair = 0
    
    self.squats = 1
    local n = 4
    self.squat = {n,-n,n,-n}
    
    self.timer:every(.5, function()
        self.body.offset_y = (self.body.offset_y)+self.squat[self.squats]
        self.body.noFixOffset = true
        self.head.offset_y = self.body.offset_y
        
        self.hair_outline.offset_y = self.body.offset_y+self.exHair
        self.hair_base.offset_y = self.body.offset_y+self.exHair
        
        self.squats = self.squats+1
        if self.squats > 4 then
            self.squats = 1
        end
    end)
    
    
    self:__draw()
    game.set_image_filter = nil
    
    self:setHair("hair4","darkblue")
end

function Person:setHair(hairName, color)
    game.set_image_filter = "nearest"
    self.hair_outline.source = string.format("person/hair/%s.png", hairName)
    if love.filesystem.exists(string.format("deadend/assets/person/hair/%sc.png", hairName)) then
        self.hair_base.source = string.format("person/hair/%sc.png", hairName)
        self.hair_base.color = color or nil
        self.hair_outline.color = nil
    else
        self.hair_base.source = "none.png"
        self.hair_outline.color = color or nil
    end
    if hairName == "hair4" then
        self.exHair = -3
    else
        self.exHair = 0
    end
    
    self:__draw()
    game.set_image_filter = nil
end

function Person:walkTo(thing)

    local x, y, w, h = thing:getRect()
    
    self.walking = x - self.x
    self.destination = {x=x,y=y,w=w,h=h,getRect=function() return thing:getRect() end, item=thing}--thing.rect
    self.legs:switch("walk")
    
end

function Person:update(dt)
    Thing.update(self, dt)
    if self.walking then
        self.vx = self:getDir(self.walking)*300--120
        self.flipX = self:getDir(self.walking)
        if self:box_collides(self.destination) then
            self.walking = nil
            self.vx = 0
            local item = self.destination.item
            if self == self.room.player then
                self.room:showActions(item)
            end
        end
        
        self.legs:switch("walk")
    else
        self.legs:switch("idle")
    end
end

function Person:draw()
    if self.destination then
        love.graphics.rectangle("fill", self.destination.x, self.destination.y, 100,100)
    end
end

function Person:getHand(other, givenAnalysis)
    local analysis = givenAnalysis or self:get("analysis")
    
    local strDiff = self:get("strength")-other:get("strength")*analysis
    local strDefenseDiff = self:get("strength")-other:get("defense")*analysis
    local healthDiff = self:get("health")-other:get("health")*analysis
    local accuracyDiff = self:get("accuracy")-other:get("agility")*analysis
    
    local val = ((strDiff+strDefenseDiff+healthDiff+accuracyDiff)/4)
    
    
    
    local strDiff = other:get("strength")*analysis-self:get("strength")
    local strDefenseDiff = other:get("strength")*analysis-self:get("defense")
    local healthDiff = other:get("health")*analysis-self:get("health")
    local accuracyDiff = other:get("accuracy")*analysis-self:get("agility")
    
    local other = ((strDiff+strDefenseDiff+healthDiff+accuracyDiff)/4)
    
    log("Hands at "..val.." to "..other)
    
    return val, other
end

function Person:getAction(other)
    if other.isPerson then
        local myHand, otherHand = self:getHand(other)
        
        if math.floor(myHand) <= math.floor(otherHand) and self:get("confidence")>=4 then
            return "suggestTeamUp"
            
        elseif myHand > otherHand*(math.random(12,15)/10) and self:get("bloodlust")>=1 then
            return "attack"
        
        elseif myHand >= otherHand then
            
        
        elseif myHand < otherHand and self:get("confidence") <= -4 then
            return "flee"
        
        elseif self:get("confidence") >= 4 and myHand > otherHand and self:get("morale")>3 then
            return "request"
            
        elseif self:get("confidence") >= 4 and myHand > otherHand then
            return "ignore"
        
        elseif self:get("confidence") >= 5 and myHand > otherHand and self:get("morale")>4 then
            return "give"
        
        elseif self:get("confidence") >= 4 and myHand > otherHand and self:get("morale")>4 then
            return "help"
        
        elseif self:get("bloodlust") >= 3 then
            return "attack"
        
        else
            return "ignore"
            
        end
    end
end

return Person