local Room = toybox.Room("GameRoom")

local ww, hh = W()/1000, H()/560
local flx, fly = 135*ww, 403*hh

rww, rhh = ww, hh

local Tile = toybox.Object("Tile")

function Tile:create(k)
    self._x, self._y = k._x, k._y
    self.static = true
    self.solid = true self.color="pink"
end

function Tile:draw()
    if toybox.debug then
        love.graphics.print(self._x..","..self._y,self.x,self.y)
    end
end

function Room:checkDoors()
    for x = 1, 4 do
        if self.openDoors[x] then
            self.doors[x]:open()
        else
            self.doors[x]:close()
        end
        
       -- if not self.entrances[x] and not self.doors[x].destroyed then
        self.doors[x].deactive = true--(true)
        if self.entrances[x] then
          if self.doors[x].destkroyed then
            self.doors[x]:add_to_room(self)
            self.doors[x].destroyed = false
          end
            self.doors[x].deactive = false
            if self.doors[x].opened then
                self.doors[x]:open()
            else
                self.doors[x]:close()
            end
        end
        
    end
end

RoomData = class:extend("RoomData")
function RoomData:__init__(kwargs)
    self.objects = {}
    self.entrances = {}
    self.x = kwargs.x
    self.y = kwargs.y
    
    self.openDoors = {false, false, false, false}
    
end

function Room:setData(roomData)
    self.cancelActs()
    
    self.x = roomData.x
    self.y = roomData.y
    
    self.entrances = roomData.entrances
    self.openDoors = roomData.openDoors
    self:checkDoors()
end

function Room:setup(kwargs)
    self.x = kwargs.x
    self.y = kwargs.y
    
    self.entrances = {}
    self.openDoors = {false, false, false, false}
    
    self.room = toybox.NewBaseObject({
        x=0, y=0, w=W(), h=H()
    })
    
    self.source = kwargs.source or "cabin"
    
    self.room:center()
    self.room.solid = false
    self.room.static = true
    self.room.depth = -5
    
    game.set_image_filter = "nearest"
   -- local g,n = love.graphics.getDefaultFilter()
   -- love.graphics.setDefaultFilter()
    self.room.background = self.room:add_image("room/room_background.png")
    self.room.background.color = kwargs.backgroundColor or {0,0,0}--"black"
    
    self.room.outline = self.room:add_image("room/room_outline.png")
    
    self.room.interior = self.room:add_image("room/cabin/cabin_interior.png")
    
    --self.room.gradient = self.room:add_image("room/room_shading_gradient.png")
    self.room.shading = self.room:add_image("room/room_shading.png")
    --self.room.s=self.room:add_image("room/room_background_door.png")
    
    self.doors = {}
    
    self.left_door = self:addDoor(69*ww, 230*hh, 52*ww, 200*hh, 93*ww, 227*hh, "left_door")
    self.left_door_2 = self:addDoor(274*ww, 191*hh, 82*ww, 200*hh, 131*ww, 206*hh, "left_door_2", 224*ww, 193*hh)
    
    self.right_door_2 = self:addDoor(606*ww, 196*hh, 82*ww, 200*hh, 131*ww, 206*hh, "right_door_2")
    self.right_door = self:addDoor(891*ww, 201*hh, 52*ww, 227*hh, 94*ww, 244*hh, "right_door",891*ww-52*ww,201*hh)
    
   -- self.left_door:open()
    
    self:__draw()
    for _, door in ipairs(self.doors) do
        door:open()
        door:__draw()
        door:close()
    end
    self:must_draw(self.room)
    self:must_update(self.room)
    
    --game.set_image_filter = nil
    
    self.background_color = {1,1,1}--colors.grey-- {0,0,0,0}
    
    toybox.make_grid(self)
    
    self:loadTiles()
    
    self.characters = {}
    self.objects = {}
    
    self:activate_gooi()
    --gooi.currentGroup = "room"
    
    local dw = W()*.8
    local dh = H()/4.5
    
    self.dialog = gooi.newPanel({
        x = 10,
        y = H()-10-dh,
        w = dw,
        h = dh,
        layout = "grid 1x6",
        groupd = "room",
    }):setColspan(1,2,5)
    
    self.portrait = gooi.newButton({
        text = ""
    })
    self.speech = gooi.newLabel({
        text = "Hmmmm, maybe here...\n...\nNah."
    }):left()
    self.speech.yOffset = 5
    self.dialog:add(self.portrait)
    self.dialog:add(self.speech)
    self.speech.opaque = true
    gooi.removeComponent(self.dialog)

    self.actionsUI = gooi.newPanel({
        x = 0,
        y = 0,
        w = W()*.25,
        h = H()*.8,
        layout = "grid 7x4"
    }):setColspan(1,2,3)
    :setColspan(2,1,4)
    :setColspan(3,1,4)
    :setColspan(4,1,4)
    :setColspan(5,1,4)
    :setColspan(6,1,4)
    :setColspan(7,1,4)
    
    self.actionsUI.opaque = true
    
    
    local function cancelActs(b)
        gooi.removeComponent(self.actionsUI)
            if self.objItem then
                self.objItem.color = self.objItem.oc
                self.objItem = nil
            end
    end
    
    self.cancelActs = cancelActs
    
    local function doAct(b)
        if b.action then
            game:playAction(b.action, self.currentObject)
            cancelActs()
        end
    end
    
    self.cancelActions = gooi.newButton({text="X"}):onRelease(cancelActs)
    self.cancelActions.fgColor = lume.copy(getColor("red"))
    self.actionObjectLabel = gooi.newLabel({text=""})
    
    self.actionsUI:add(self.cancelActions)
    self.actionsUI:add(self.actionObjectLabel)
    self.actionButtons = {}
    
    self.maxActions = 6
    for x = 1, self.maxActions do
        local b = gooi.newButton({text="??"}):onRelease(doAct)
        self.actionButtons[#self.actionButtons+1] = b
        self.actionsUI:add(b)
    end
    
    cancelActs()
    
    toybox.debjug = 1
    
    self.camera.scale = 1
    self:spawn("table", self:getTile(2,1))--"space")

end

function isThing(item)
    return item.isThing and item~=toybox.room.player and not item.deactive
end

function Room:mousepressed(x,y)
    local items, len = self.world:queryRect(x,y,5,5,isThing)
    
    local item = items[1]
    if item and item ~= self.objItem then
        
        if self.player then
            self.player:walkTo(item)
            
            if self.objItem then
                self.objItem.color = self.objItem.oc
                self.objItem = nil
            end
            
            self.cancelActs()
            self.objItem = item
            item.oc = item.color
            item.color = "yellow"
            self.player.name_tag = (item.name or item.class.name)..","..math.random()
        end
    end
    log(self.room.x..","..self.room.y)
    
    self.background_color = {0,0,0}
end

local function rect(door)
    return door.world:getRect(door)
end
local function closeDoor(door)
    door.source = string.format("room/%s/doors/%s.png",door.room.source,door.ttype)
    door:set_box(door.clx, door.cly, door.clw, door.clh)
    door.depth = DEPTHS.DOORS_CLOSED
    door.closed = true
    door.room.openDoors[door.num] = false
    door.opened = false
    
end

local function openDoor(door)
    door.source = string.format("room/%s/doors/%s_open.png",door.room.source,door.ttype)
    door:set_box(door.opx, door.opy, door.opw, door.oph)
    door.depth = (door == door.room.right_door or door == door.room.left_door) and DEPTHS.DOORS
                 or DEPTHS.DOORS_CLOSED
    door.closed = false
    door.room.openDoors[door.num] = true
    door.opened = true
    if door.room.player then door.room.player.name_tag = "yoh"..math.random() end
    door.room.camera:shake(15,.2,30,"x")
    if door.room.entrances[door.num] then
        door.room.entrances[door.num].openDoors[reverseDirs[door.num]] = true
    end
end

function dUpdate(d)
    if d.opened then
        d:open()
    else
        d:close()
    end
end

local xxes = {147.4, 446, 704, 1072}

function Room:addDoor(x, y, w, h, open_w, open_h, source, open_x, open_y)
    local d  = --Thing({--
    toybox.NewBaseObject({
        x = 0, y = 0,
        w = W(), h = H()
    })
    
    d.getRect = rect
    d.debug=1
    
    d.update = null
    d.isDoor = true
    d:center()
    
    d.isDoor = true
    d.isThing = true
    
    d.solid = false
    d.isDoor = true
    d.static = true
    d.depth = DEPTHS.DOORS
    
    d.clx, d.cly = x, y
    d.clw, d.clh = w, h
    
    d.opx, d.opy = open_x or x, open_y or y
    d.opw, d.oph = open_w, open_h
    
    d.ttype = source
    
    d.close = closeDoor
    d.open  = openDoor
    
    --d.update = dUpdate
    
    self.doors[#self.doors+1] = d
    
    d.num = #self.doors
    d.xx = xxes[d.num]
    
    d:close()
    
    return d
end

function Room:draw_instance(x)
    game.set_image_filter = "nearest"
    if x.deactive then
        return
    else
        x:__draw()
    end
end

function Room:loadTiles()

    for x = 1, 9*3 do
        local flx = flx
        local tw, th, thh
        if x <= 9 then
            tw = 82*ww
            th = 54*hh
            thh = 0
        elseif x <= 18 then
            tw = 92*ww
            th = 54*hh
            flx = flx-((tw*9)-(82*ww*9))/2
            thh = 56
        else
            tw = 102*ww
            th = 54*hh
            flx = flx-((tw*9)-(82*ww*9))/2
            thh = 56*2
        end
        
        local t = Tile:new({
            x = flx+tw*(x%9),
            y = fly+thh*hh,
            w = tw,
            h = th,
            _x = x%9,
            _y = 1-(x>18 and 2 or x>9 and 1 or 0)
        })
        self:storeTile(t)
    end
    
    for y = 1, 7 do
        for x = 1, 12 do
            local tw, th = 82*ww, 54*ww
            local t = Tile:new({
                x = flx + tw * (x-1) - 82*ww,
                y = fly - 56 *hh * y - 10,
                w = tw,
                h = th,
                _x = x,
                _y = y+1,
            })
            t.solid = false
            self:storeTile(t)
        end
    end


end

function Room:setActions(list)
    for x = 1, self.maxActions do
        local b = self.actionButtons[x]
        local act = list[x]
        if act then
            b.action = act
            b:setText(act.name)
            b.opaque = true
        else
            b:setText("")
            b.opaque = false
        end
    end
end

function Room:showActions(obj)
    local list = Thing.getActions(obj)
    self.currentObject = obj
    self:setActions(list)
    gooi.addComponent(self.actionsUI)
    self.actionObjectLabel:setText(obj.name or obj.class.name)
end

local t_update = function(t1, t2)
    local t = {}
    for x, i in pairs(t1) do
        t[x] = i
    end
    for x, i in pairs(t2) do
        t[x] = i
    end
    return t
end

function Room:spawn(dat, x, y)
    x = x or "space"
    if type(x) == "string" then
        x = self:getPosition(x)
    end
    
    if type(x) == "table" and x.isTile then
        y = x.y
        x = x.x
    end
    
    local item = objects[dat] or dat
    local itemm = (dat.isItem and Item or Setpiece):new(t_update(item, {
        x = x,
        y = y,
        h = (item.h or 50)*hh,
        w = (item.w or 50)*ww
    }))
    
    if item.gravity then
        itemm.gravity = type(item.gravity) == "number" and item.gravity or 10
    end
    
    if item.offset_y then
        itemm.offset_y = itemm.offset_y + item.offset_y
    end
    
    
end
    

function Room:keypressed()
    self.camera.scale = self.camera.scale == 1 and .3 or 1
    Party(self,nil,3,"background_color")--.color = "blue"
end

function Room:step(dt)
    self.room:move_to(0,0)
end

function Room:draw_after()
    gooi.draw(gooi.currentGroup)
    self.talker = self.player
    if self.talker then
        local pt = self.talker
        pt:redraw(self.portrait.x, self.portrait.y+self.portrait.h*0, nil, pt.w*2, pt.h*2)
    end
end

return Room