
require('toybox.utils.rgba')
require('toybox.utils.round')
require('toybox.utils.sign')
require('toybox.utils.spairs')

LGML = setmetatable({}, {
  __call = function (t, options)
    t.__entry = require(options.entry)
    t.__debug = options.debug or false
  end
})

toybox = LGML


class = require "toybox.libs.middleclass"

Andralog = require "toybox.libs.andralog"

Paddy = require "toybox.libs.paddy"
Chrono = require "toybox.libs.chrono"
Camera = require "toybox.libs.camera"

lume = require "toybox.libs.lume"
inspect = require "toybox.libs.inspect"
require  "toybox.libs.tools"
aspect = require "toybox.libs.aspect"
bump = require "toybox.libs.bump"
animx = require "toybox.libs.animx"

ProFi = require "toybox.libs.ProFi"

JumperGrid = require "toybox.libs.jumper.grid"
PathFinder = require "toybox.libs.jumper.pathfinder"
    
    
local oor = require

old_require = oor
old_requiren = function(n)
    return oor(string.format("libs.steering%s",n))
end

require = function(n)
    return oor(string.format("toybox.libs.steering.%s",n))
end

SteeringAgent = require "src/agent/steeringAgent"

require = old_require

luastar = require_and_sign "toybox.libs.luastar"

colors = require_and_sign "toybox.libs.color"

aspect = require_and_sign 'toybox.libs.aspect'
saveTable = require "toybox.libs.saveTable"

require "toybox.libs.gooi"

toybox.make_grid = require "toybox.utils.grid"
toybox.make_boid = require "toybox.utils.boid"

toybox.bars = require "toybox.extra.bars"

print = log

InputManager = require "toybox.managers.InputManager"
AudioManager = require_and_sign "toybox.managers.AudioManager"
media = require_and_sign "toybox.utils.media"

function getRandom(r)
    if type(r) == "table" then
        if type(r[1])=="number" and #r==2 then
            return math.random(r[1], r[2])
        end
        return lume.randomchoice(r)
    end
    return r
end

getValue = getRandom
toybox.getRandom = getRandom
toybox.getValue = getRandom

aspect.setGame(_W,_H)

res = {}
res.beginRendering = function()
    aspect.start()--push:apply("start")
end
    
res.endRendering = function()
    aspect.stop()--push:apply("end")
end
    
function res.update(dt)
    aspect.update(dt)
end

lmx = love.mouse.getX
lmy = love.mouse.getY
    
res.getMousePosition = function(w,h,x,y)
    local x,y = x or w or lmx(), y or h or lmy()
    return aspect.toGame(x,y)
end
    
love.mouse.getX = function()
    local x,y = res.getMousePosition()
    return x
end
    
love.mouse.getY = function()
    local x,y = res.getMousePosition()
    return y
end

local luu = love.update
function love.update(dt)
    if lui then luu(dt) end
    animx.update(dt)
    res.update(dt)
end
lu=love.update
Object = require "toybox.entities.Object"
Room = require "toybox.entities.Room"
Sprite = require "toybox.entities.Sprite"
Controller = require "toybox.entities.Controller"
Game = require "toybox.entities.Game"

    
    IM = InputManager:new()
    input = IM
    inputManager = IM
    IM.active = true


local SETTINGS_FILE = "settings.dat"

local function getSettingsToLoad()
    if love.filesystem.isFile(SETTINGS_FILE) then
        return love.filesystem.load(SETTINGS_FILE, "r")()
    end
end

local function getSettingsToSave(data)
    local tmp = love.filesystem.newFile(string.format("%s_TMP",SETTINGS_FILE),"w")
    tmp:write(data)
    local value = love.filesystem.load(string.format("%s_TMP",SETTINGS_FILE), "r")()
    if type(value) == "table" then
        local file = love.filesystem.newFile(SETTINGS_FILE, "w")
        file:write(data)
        return true
    elseif nil then--game then--.map then
        GTimer:after(3,function() gooi.alert({
            text = "No space to save data!"
        })
        end)
    end
    
end

LGML.getColor = function(col,alpha)
    if type(col) == "table" then 
        if alpha then
            col = {col[1],col[2],col[3]}
            col[4] = alpha
        end
        return col
    end
    local col = colors[col] or colors.red
    col = {col[1],col[2],col[3],col[4]}
    if alpha then
        col[4] = alpha
    end
    return col
end

LGML._data = {}
LGML.getData = function(name)
    SETTINGS_FILE = name or "settings.dat"
    local d = LGML._data[name] or getSettingsToLoad() or {}
    LGML._data[name] = d
    return d
end

function LGML.saveData(name)
    SETTINGS_FILE = name or "settings.dat"
    getSettingsToSave(saveTable(LGML._data[name] or getSettingsToLoad()))
end


LGML.deactivate_gooi = function()
    gooi_activated = false
end

LGML.Game = function (name)
  return require('toybox.entities.Game'):subclass(name)
end

LGML.Room   = function (name) 
  return require('toybox.entities.Room'):subclass(name)
end

LGML.Object = function (name)
  return require('toybox.entities.Object'):subclass(name)
end

LGML.Sprite = function (name)
  return require('toybox.entities.Sprite'):subclass(name)
end

LGML.new_sprite = function (...)
  return require('toybox.entities.Sprite'):new(...)--subclass(name)
end

LGML.Controller = function (name)
  return require('toybox.entities.Controller'):subclass(name)
end

LGML.BaseController = function (name)
  return require('toybox.entities.BaseController'):subclass(name)
end

LGML.BaseObject = function (...)
  return require('toybox.extra.BaseObject'):subclass(...)
end

LGML.NewBaseObject = function (...)
  return require('toybox.extra.BaseObject'):new(...)
end

LGML.Projectile = function (name)
  return require('toybox.extra.Projectile'):subclass(name)
end

LGML.NewProjectile = function (...)
  return require('toybox.extra.Projectile'):new(...)
end

LGML.LifeBar = function(parent, kwargs)
    if kwargs and parent then
        kwargs.parent = parent
    end
    
    kwargs = kwargs or parent
    
    return toybox.bars.LifeBar:new(kwargs)
end

LGML.ProgressBar = function(parent, kwargs)
    if kwargs and parent then
        kwargs.parent = parent
    end
    
    kwargs = kwargs or parent
    
    return toybox.bars.ProgressionBar:new(kwargs)
end

function love.load()
  LGML.__instance = LGML.__entry:new()
  _G.game = LGML.__instance
end

function love.update(dt)
  lu(dt)
  LGML.__instance:__step(dt)
end

function love.draw()
  LGML.__instance:__draw()
end

function love.mousepressed(...)
  LGML.__instance:mousepressed(...)
end

function love.mousereleased(...)
  LGML.__instance:mousereleased(...)
end

function love.mousemoved(...)
  LGML.__instance:mousemoved(...)
end

function love.keypressed(...)
  LGML.__instance:keypressed(...)
end

function love.keyreleased(...)
  LGML.__instance:keyreleased(...)
end

function love.textinput(...)
  LGML.__instance:textinput(...)
end

function love.touchpressed(...)
  LGML.__instance:touchpressed(...)
end

function love.touchreleased(...)
  LGML.__instance:touchreleased(...)
end

function love.touchmoved(...)
  LGML.__instance:touchmoved(...)
end
    
function love.focus(focus)
    if not LGML.__instance then return end
    
    local game = LGML.__instance
    
    if game.room and game.room._focus then
        game.room:_focus(focus)
    end
    
    if not focus then
        game.stopped_focus = true
        if media.music then
            media.music:pause()
        end
    elseif game.stopped_focus then
        game.stopped_focus = false
        if media.music then
            media.music:play()
        end
    end
end

return LGML