-- Create class
local Room = class('Room')

-- Room:initialize
-- Room constructor
function Room:initialize(p)
  local old_room = toybox.room
  toybox.room = self
  -- Room creation
  self.objects = {}
  self.instances = {}
  self.viewports = {}
  self.viewport = false
  self.viewport_target = false
  self.width = W()--640
  self.height =H()-- 480
  
  self.physics_scale = 1
  self.gravity = 0
  self.world = bump.newWorld(16)
  
  self.timer = Chrono()
  self.controllers = {}
  
  self._must_draw = {}
  self._must_update = {}
  
  self.name = "room"
  self.total_steps = 0

  self:add_viewport('default', {})
  self:set_viewport('default')
  
  self.noRefresh = p.noRefresh
  self._instanceRoom = true
  
  self.freeze = p.freeze or .1
  
  self._squashing = {}
 
  -- Execute setup
  if type(self.setup) == 'function' then
    self:setup(p)
  end
  
  toybox.room = old_room or self
end
local log= tostring
-- Room:update
-- Update instances accordingly its update method
function Room:__step(dt)
  log("freeze?")
  if type(self.freeze) == "number" and self.freeze > 0 then
    self.freeze = self.freeze - dt
    return
  elseif self.freeze == true then
    return
  end
  log("up step")
  self.total_steps = self.total_steps + 1
  
  self:__update_controllers(dt) log("con done")
  self:__update_viewport(dt) log("cam done")
  self:__update_instances(dt) log("ins done")
  
  self.timer:update(dt) log("timer done")

  self:__update_squashing()
  
  if type(self.step) == 'function' then
    self:step(dt)
  end
  
  if type(self.update) == 'function' then
    self:update(dt)
  end
  
  log("others")
  if self.gooi_activated then
    gooi.update(dt)
  end
  log("update steo done")
  
end
-- Room:draw
-- Update instances accordingly its draw method
function Room:__draw()
  -- Room drawing
  if self.background_color then
    love.graphics.setBackgroundColor(getColor(self.background_color))
  end
 
  if self.gooi_activated and self.draw_gooi then
    gooi.draw(gooi.currentGroup)
  end
  
    if type(self.draw_before) == 'function' then
      self:draw_before()
    end
  if self.viewport then
    self.viewport:attach()

    -- Draw inside viewport
    self:__draw_instances()

    if type(self.draw) == 'function' then
      self:draw()
    end

    self.viewport:detach()

    -- Added for library features support
    self.viewport:draw()
    
    if self.gooi_after_draw then
        gooi.draw(gooi.currentGroup)
    end
    
  else
    -- Fallback
    self:__draw_instances()

    if type(self.draw) == 'function' then
      self:draw()
    end
  end
  
  
  self:__draw_controllers()
  
  if type(self.draw_after) == 'function' then
    self:draw_after()
  end
    
  if type(self.after_draw) == 'function' then
    self:after_draw()
  end
    
    
  if type(self.post_draw) == 'function' then
    self:post_draw()
  end
end

function Room:activate_gooi(noDraw)
  self.gooi_activated = true
  self.draw_gooi = not noDraw
end

function Room:deactivate_gooi()
  self.gooi_activated = false
end

-- Room:init
-- Start executing the room instances
function Room:init()
  if (not self.noRefresh) or not self._created then
   -- self:kill() -- Reset room
    self:set_viewport('default') -- Set default viewport
    self:__create_instances() -- create instances
    self._created = true
    if self.oninit then
      self:oninit()
    end
  end
end

-- Room:kill
-- Reset room data
function Room:kill()
  self:__reset_viewport()
  self:__destroy_instances()
  
  self.controllers = {}
  
  local trash = self.onkill and self:onkill()
end

function Room:add_controller(c)
  c.id = lume.uuid()
  self.controllers[c.id] = c
end

function Room:remove_controller(c)
  self.controllers[c.id] = nil
end

function Room:tween(...)
  return self.timer:tween(...)
end

function Room:every(...)
  return self.timer:every(...)
end

function Room:after(...)
  return self.timer:after(...)
end

function Room:set_physics_scale(scale)
  self.physics_scale = scale
end

function Room:get_gravity()
    return self.gravity
end

function Room:get_height()
  return math.max(self.height, love.graphics.getHeight())
end

function Room:get_width()
  return math.max(self.width, love.graphics.getWidth())
end

-- Room:add_viewport
-- Add a viewport to the room viewport list
function Room:add_viewport(name, options)
  if not self.viewports[name] then
    self.viewports[name] = Camera(unpack(options))
  else
    print('[Room:add_viewport] Viewport already added: ', name)
  end
end

-- Room:__reset_viewport
-- Set viewport to its initial state
function Room:__reset_viewport()
  self.viewport = false
end

-- Room:set_viewport
-- set current app viewport
function Room:set_viewport(name)
  local viewport = self.viewports[name]

  if viewport then
    self.viewport = viewport
    self.camera   = viewport
    print('[Room:add_viewport] Viewport set: ' .. name)
  else
    self:__reset_viewport()
    print('[Room:add_viewport] Viewport not found: ' .. name)
  end
end

-- Room:getViewport
-- Exposes viewport
function Room:getViewport()
  return self.viewport
end

-- Room:getViewport
-- Exposes viewport
function Room:get_viewport()
  return self.viewport
end

-- Room:__update_viewport
-- Execute update method safely
function Room:__update_viewport()
  local dt = love.timer.getDelta()

  if self.viewport then
    self.viewport:update(dt)

    if self.viewport_target then
        -- Viewport target
      local xTarget = self.viewport_target.x
      local yTarget = self.viewport_target.y

      local roomWidth = self:get_width()
      local roomHeight = self:get_height()

      local halfScreenWidth = love.graphics.getWidth() / 2
      local halfScreenHeight = love.graphics.getHeight() / 2

      if xTarget < halfScreenWidth then
        xTarget = halfScreenWidth
      end

      if yTarget < halfScreenHeight then
        yTarget = halfScreenHeight
      end

      if xTarget > roomWidth - halfScreenWidth then
        xTarget = roomWidth - halfScreenWidth
      end

      if yTarget > roomHeight - halfScreenHeight then
        yTarget = roomHeight - halfScreenHeight
      end

      xTarget = math.round(xTarget)
      yTarget = math.round(yTarget)
        -- Viewport target
      local xTarget = self.viewport_target.x
      local yTarget = self.viewport_target.y

      self.viewport:follow(xTarget, yTarget)
    end
  end
end

-- Room:place_object
-- Place objects in room for giving coordinates
function Room:place_object(ObjDefinition, x, y)
  if ObjDefinition and ObjDefinition:isSubclassOf(Object) then
    table.insert(self.objects, { ObjDefinition, x, y, lume.uuid() })
  else
    print('Object is not instance of Object')
  end
end

-- Room:place_instance
-- Place already made instances
function Room:place_instance(instance, x, y)
  if x then
    instance:move_to(x,y)
    instance.x, instance.y = x,y
    
  end
  
  
  instance.room = self
  
  self:store_instance(instance)
end

-- Room:set_viewport_target
-- Set instance to be followed by viewport
function Room:set_viewport_target(instance)
  self.viewport_target = instance
end

-- Room:set_target
-- Set instance to be followed by viewport
function Room:set_target(instance)
  self.viewport_target = instance
end


-- Room:create_instanceFromObject
-- Create single instance from object item from objects array
function Room:create_instance(ObjDefinition, x, y, id)
  if ObjDefinition and ObjDefinition:isSubclassOf(Object) then
    local ninstances = table.getn(self.instances)
    local id = id or lume.uuid()-- ninstances + 1000
    log("created")

    local instance = ObjDefinition:new({
      id = id,
      x = x,
      y = y,
      _created = true,
      room = self
    })
    
    
    instance._created = true

    self:store_instance(instance)
  end
end

function Room:store_instance(instance)
  instance.id = instance.id or lume.uuid()
  instance.room = self
  self.instances[instance.id] = instance
end

function Room:has_instance(instance)
    instance.id = instance.id or lume.uuid()
    return self.instances[instance.id]
end

function Room:squash(obj, w, h, time, func, after, centered)
    local time = time or 1
    local timer = obj.timer or self.timer
    
    if type(after) == "table" then
        local n = after
        local w = type(n[1]) == "number" and n[1] or type(n[2]) == "number" and n[2] or n.w or 1
        local h = type(n[2]) == "number" and n[2] or type(n[3]) == "number" and n[3] or n.t or w
        local t = type(n[1]) == "string" and n[1] or type(n[2]) == "string" and n[2] or
            type(n[3]) == "string" and n[3] or n.func or "in-bounce"
        after = function()
            self[(centered and "squash_centered" or "squash")](self, obj, w, w, h, t)
        end
    end
    
    obj._sqow = obj._sqow or obj.w
    obj._sqoh = obj._sqoh or obj.h
    
    if centered then
        obj._sqox = obj._sqox or obj.ogx or obj.x
        obj._sqoy = obj._sqoy or obj.ogy or obj.y
    end
    
    obj.squashing = (w+h)/2--obj.squashing or 0
    obj.doneSquash = (obj.doneSquash or 0)+1
    local ow, oh = obj._sqow, obj._sqoh
    timer:tween(time, obj, {
        w = ow*w,
        h = oh*h,
        doneSquash = obj.doneSquash-1,
        --squashing = (w+h)/2
    }, func or "out-quad", after)--,function()
    
    if centered then
        self._squashing[obj] = true
    end
end

function Room:squashW(obj, w, time, func, after, tt,f2)
    if type(func)  == "number" then
        local ww = func
        tt = after
        f2 = tt
    end
    if type(after) == "number" then
        
        after = function()
            self:squashW(obj, ww, tt or time,f2)
        end
    end
    
    return self:squash(obj, w, 1/w, time, func, after)
end

function Room:squash_centered(obj, w, h, time, func, after)
    
    self:squash(obj, w, h, time, func, after, true)
    
end

function Room:squash_ui(ui)
    local obj, centered = ui, true
    
    if ui.object then
        obj = ui.object
        centered = false
    end
    
    if ( obj.doneSquash ) <= .1 then
        self:squash(obj, .5, 1.2, .3, nil, {1,.1}, centered)
    end
end

function Room:tween_in_ui(ui, time, func, after)
    if not (ui.ogx or ui.ogy) then
        return
    end
    
    ui.tweening = 1
    self:tween(
        time or 1,
        ui,
        {x=ui.ogx or ui.x, y=ui.ogy or ui.y, tweening=-1},
        func or "in-bounce",
        after
    )
end

function Room:tween_out_ui(ui, time, func, after)
    if not (ui.outx or ui.outy) then
        return
    end
    
    ui.tweening = 1
    self:tween(
        time or 1,
        ui,
        {x=ui.outx or ui.x, y=ui.outy or ui.y, tweening=-1},
        func or "out-quad",
        after
    )
end

function Room:__update_squashing(dt)
    for obj, _ in pairs(self._squashing) do
        obj.x = obj._sqox-(obj.w-obj._sqow)/2
        obj.y = obj._sqoy-(obj.h-obj._sqoh)/2
        
        if obj.doneSquash <= 0 then
            self._squashing[obj] = nil
        end
    end
end

-- Room:__create_instances
-- Loop though all objects and instantiate them
function Room:__create_instances()
  for index, object in ipairs(self.objects) do
    self:create_instance(unpack(object))
  end

  print('[Room:createIntances] Creating instances for: ' .. Room.name)
end

-- Room:destroy instances
-- Reset instances table
function Room:__destroy_instances(all)
  for index in pairs(self.instances) do
    local item = self.instances[index]
    
    if item and (item._created or all) then
      self.world:remove(item)
      self.instances[item.id] = nil
    end
  end
end

function Room:destroy_instance(instance)
  self.instances[instance.id] = nil
  self.world:remove(instance)
end

function Room:destroy_all_instances()
  return self:__destroy_instances(true)
end

function Room:must_update(obj, remove)
    self._must_update[tostring(obj)] = (not remove and obj) or nil
end

function Room:must_draw(obj, remove)
    self._must_draw[tostring(obj)] = (not remove and obj) or nil
end

-- Depth sorting method
function Room.depth_sorter(a, b)
  return b.depth > a.depth
end

-- Room:__draw_instances
-- loop though instances and draw them
function Room:__draw_instances()
  local obj
  local x, y, w, h = self:get_viewport():getVisible()--Window()
  
  local visibleThings, len = self.world:queryRect(x,y,w,h)
    
  for x, i in pairs(self._must_draw) do
      visibleThings[len+1] = i
      len = len+1
  end
  
  if 1 then table.sort(visibleThings, self.depth_sorter) end
  mm=0
  for i,m in pairs(self.instances) do
      mm=mm+1
     end
  --assert(len>0, mm.."?")
  for i = 1, len do
    obj = visibleThings[i]
    local cam
    if obj.no_camera and self:get_viewport().attached then
        self:get_viewport():detach()
        cam = true
    end
    self:draw_instance(obj)
    if cam then
        self:get_viewport():attach()
    end
  end
end

function Room:draw_instance(obj)
  obj:__draw()
end

-- Room:__update_instances
-- loop though instances and update them
function Room:__update_instances(dt)
  log("ins?")
  if self.classic then
  local obj log("here")
  local x, y, w, h = self:get_viewport():getVisible()--Window()
  log("na"..x..","..y..","..w..","..h)
  --x = math.floor(x) y = math.floor(y)
  if math.abs(x)>1.1e+13 then
    x = getDir(x)*99999
  end
  if math.abs(y)>1.1e+13 then
    y = getDir(y)*99999
  end
  local vw = (self.vw or w)*(self.multipleW or 1)
  local vh = (self.vh or h)*(self.multipleH or 1)
  local visibleThings, len = self.world:queryRect(x-(vw-w)/2,y-(vh-h)/2,
  vw,vh)

 
  if 1 then table.sort(visibleThings, Room.depth_sorter) end
  mm=0
  for i,m in pairs(self.instances) do
      mm=mm+1
     end
  --assert(len>0, mm.."?")
  for i = 1, len do
    obj = visibleThings[i]
    if not self._must_update[tostring(obj)] then
      self:__update_instance(obj, dt)
    end
  end
  
  for x, i in pairs(self._must_update) do
    self:__update_instance(i, dt)
  end
  else
  for index, instance in pairs(self.instances) do
    self:__update_instance(instance, dt)
  end
  end
end

function Room:__update_instance(instance,dt)
  instance:__step(dt)
end

function Room:__update_controllers(dt)
  for index, controller in pairs(self.controllers) do
    controller:__step(dt)
  end
end

function Room:__draw_controllers()
  for index, controller in pairs(self.controllers) do
    controller:__draw()
  end
end

function Room:__keypressed(key)
  if self.keypressed then
     self:keypressed(key)
  end

  if self.gooi_activated then
    gooi.keypressed(key)
  end
  
  for index, controller in pairs(self.controllers) do
   -- controller:__keypressed(key)
  end
end

function Room:__keyreleased(key)

  if self.gooi_activated then
    gooi.keyreleased(key)
  end
  
  for index, controller in pairs(self.controllers) do
   -- controller:__keyreleased(key)
  end
end

function Room:__textinput(text)
  if self.typed then return end
  self.typed = true
  if self.gooi_activated then
    gooi.textinput(text)
  end
end

function Room:__mousepressed(x,y,button)
  if self.gooi_activated then
    gooi.pressed()
  end
  
  for index, controller in pairs(self.controllers) do
    controller:__mousepressed(x,y,button)
  end
  
  if self.mousepressed then
    return self:mousepressed(x,y,button)
  end
end

function Room:__mousereleased(x,y,button)
  if self.gooi_activated then
    gooi.released()
  end
  
  for index, controller in pairs(self.controllers) do
    controller:__mousereleased(dt)
  end
  
  if self.mousereleased then
    return self:mousereleased(x,y,button)
  end
end

function Room:__mousemoved(x,y)
  if self.gooi_activated then
    gooi.moved(x,y)
  end
  
  for index, controller in pairs(self.controllers) do
    controller:__mousemoved(x,y,button)
  end
  
  if self.mousemoved then
    return self:mousemoved(x,y,button)
  end
end

function Room:touchpressed(id, x, y, dx, dy, pressure)
  for index, controller in pairs(self.controllers) do
    controller:touchpressed(id, x, y, dx, dy, pressure)
  end
end

function Room:touchreleased(id, x, y, dx, dy, pressure)
  for index, controller in pairs(self.controllers) do
    controller:touchreleased(id, x, y, dx, dy, pressure)
  end
end

function Room:touchmoved(id, x, y, dx, dy, pressure)
  for index, controller in pairs(self.controllers) do
    controller:touchmoved(id, x, y, dx, dy, pressure)
  end
end

return Room