local Sprite = class:extend("Sprite")

function Sprite:__init__(p,kwargs)
    self.uuid = lume.uuid
    
    if not kwargs then
        kwargs = p
        p = nil
    end
    
    self.parent = (kwargs and p) or kwargs.parent or {}
    
    --assert(self.parent and self.parent.isObject, "<Sprite> class needs an <Object> parent, ya Dingdong.")
    
    self.x = kwargs.x or self.parent.x or 0
    self.y = kwargs.y or self.parent.y or 0
    
    self.w = kwargs.w or self.parent.w or 16
    self.h = kwargs.h or self.parent.h or 16
    
    self.offset_x = kwargs.offset_x or self.parent.offset_x or 0
    self.offset_y = kwargs.offset_y or self.parent.pffset_y or 0
    
    self.angle = kwargs.angle or 0
    self.map = kwargs.map or toybox.room
    
    self.noResize = kwargs.noResize
    self.noFixOffset = kwargs.noFixOffset
    
    self.bind = kwargs.bind ~= false
    --error(tostring(llkk==game) ..traceObject(game))--(game,3))
    
    local anim = kwargs.animations and lume.randomchoice(kwargs.animations) or kwargs
    self.source = game:getAsset(anim.source..(anim.useImages and "/0.png" or ""))
    self.animation = animx.newActor()
    
    self.kwargs = kwargs
    
    if kwargs.source then
    self.animation:
        addAnimation(kwargs.name or "norm",{
            img = kwargs.source,
            delay = kwargs.delay or .2,
            useImages = kwargs.useImages and true,
            mode = kwargs.mode or "once",
            onAnimOver = kwargs.onAnimOver or kwargs.onAnimEnd or kwargs.onAnimFinish or kwargs.destroy and function()
                self.parent:destroy()
            end,
        }):switch(kwargs.name or "norm")
    end
    
    for x , i in pairs(kwargs.animations or {}) do
        self.animation:addAnimation(i.name or x, {
            img =i.source,
            delay = i.delay or kwargs.delay or .2,
            useImages = i.useImages and true,
            mode = i.mode or kwargs.mode or "loop",
            noOfFrames = i.noOfFrames,
            qw = i.qw,
            qh = i.qh,
            frames= i.frames,
            spritesPerRow = i.spr or i.spritesPerRow,
            onAnimOver = function()
                local ii = i.onAnimOver or i.onAnimEnd or i.onAnimFinish or kwargs.onAnimOver or kwargs.onAnimEnd or kwargs.onAnimFinish
                if ii then
                    return ii(self)
                end
            end
        }):switch(i.name or x)
    end
    
    self:update(0)
end

function Sprite:switch(s)
    self.animation:switch(s)
    return self
end

function Sprite:getSource()
    return self.source
end

function Sprite:getCurrentImage()
    local a = self.animation:getCurrentAnimation()
    return a.useImages and a:getCurrentQuad() or a:getTexture()
end

function Sprite:update(dt) self.upp=true
    if self.bind then
        self.x, self.y = self.parent.x, self.parent.y
        if not self.noResize then
            self.w, self.h = self.parent.w, self.parent.h
        end
    end
    if not self.noFixOffset then
        self.offset_x = self.parent.offset_x or 0
        self.offset_y = self.parent.offset_y or 0
    end
    self.angle = self.parent.angle or 0
    self.flipX = self.parent.flipX
    self.flipY = self.parent.flipY
    self._w, self._h = resizeImage(self.source, self.w, self.h)
end


function Sprite:draw()
    local r,g,b,a = love.graphics.getColor()
    if self.color then
        love.graphics.setColor(getColor(self.color))
    end
    
    local img = self.source assert(self.upp, "Sprite has not been updated, yo.")
    --self:update(0)
    if self.noResfize then
        self._w, self._h = self.parent._w, self.parent._h
    else
        self._w, self._h = resizeImage(img, self.w, self.h)
    end
    
    if self.parent.angle~= 0 and toybox.debug then self.dd=self.dd or self.parent.angle self.parent.color=getColor("yellow") log(math.deg(self.parent.angle)) end 
    if not (self.parent.angle) then error( inspect(self.kwargs,10)) end
    local ang = math.rad(self.parent.angle)
    if self.flipX == -1 and ang~=0 then
        ang = math.rad(180-math.deg(ang))*-1
    end
    
    self.animation:draw(
        self.x+self.offset_x+self.parent.shake_x+(self.flipX == -1 and -self.w*0 or 0),
        self.y+self.offset_y+self.parent.shake_y+(self.flipY == -1 and self.h*0 or 0) ,
        -ang,
        self._w*self.flipX, self._h*self.flipY
        ,img:getWidth()/2,--*self.flipX,
        img:getHeight()/2
    )
    love.graphics.setColor(r,g,b,a)
end

return Sprite